"""
A utils for interacting with Reddit Gateway API (Web API), w/ pushshift historical posts support.

[![Latest Version](https://img.shields.io/pypi/v/gatered.svg)](https://pypi.python.org/pypi/gatered)
[![Supported Python Versions](https://img.shields.io/pypi/pyversions/gatered)](https://pypi.python.org/pypi/gatered)
[![CI](https://github.com/countertek/gatered/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/countertek/gatered/actions/workflows/ci.yml)
[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)
[![GitHub license](https://img.shields.io/github/license/countertek/gatered)](https://github.com/countertek/gatered/blob/main/LICENSE)

## Why this library?

Although Reddit has developer APIs and there are existing libraries (e.g. praw) to interact with reddit, there are still several drawbacks in terms of collecting data:

- An API key is needed to collect data.
- Rate limit is based on API keys.
- Some fields are missing using developer APIs.

Therefore, **gatered** exists just to counter this problem. It directly access Reddit's web API to get the whole information. No authentication is needed, and it supports proxy provided by [httpx](https://www.python-httpx.org/advanced/#http-proxying).

## Install

You can install this library easily from pypi:

```bash
# with pip
pip install gatered

# with poetry
poetry add gatered
```

## Using

The library provides easy functions to get start fast:
- `gatered.func.get_post_comments`
- `gatered.func.get_posts`
- `gatered.func.get_comments`
- `gatered.func.get_pushshift_posts`

Alternatively you can directly use `gatered.client.Client` and `gatered.pushshift.PushShiftAPI` classes as your base to implement your own logics.

Errors can be handled by importing either `gatered.RequestError` or `gatered.HTTPStatusError`, see [httpx exceptions](https://www.python-httpx.org/exceptions/) to learn more.

See [`examples/`](https://github.com/countertek/gatered/tree/main/examples/) for more examples.
"""

__version__ = "1.1.0"

# Import error classes from httpx: https://www.python-httpx.org/exceptions/
from httpx import RequestError, HTTPStatusError

from .client import Client
from .pushshift import PushShiftAPI
from .func import get_post_comments, get_posts, get_comments, get_pushshift_posts



"""
This module contains easy functions to acquire information from Reddit. 
All functions are async and can be called directly without context manager.
"""
from typing import Any, Dict, Optional
from functools import partial
from asyncio import sleep
from aiometer import run_all
import logging

from gatered.client import Client
from gatered.pushshift import PushShiftAPI
from gatered.utils import datetime, get_timestamp

log = logging.getLogger(__name__)


async def get_post_comments(
    submission_id: str,
    all_comments: bool = False,
    httpx_options: Dict[str, Any] = {},
):
    """
    Helper function to get submission and its comments.

    Provide `submission_id` (starts with `t3_`) and
    if `all_comments` is `True`, it will fetch all the comments that are nested by reddit.

    Returns a dict with `post` as dict and its `comments` as list.
    """
    async with Client(**httpx_options) as client:
        return await client.get_post_comments(submission_id, all_comments=all_comments)


async def get_posts(
    subreddit_name: str,
    sort: Optional[str] = "hot",
    t: Optional[str] = "day",
    page_limit: Optional[int] = 4,
    req_delay: int = 0.5,
    httpx_options: Dict[str, Any] = {},
):
    """
    Async Generator to get submissions batch by batch.

    Returns an async generator that yields a list of posts. Use `async for` loop to handle the results.

    - `subreddit_name` (str):
        The Subreddit name.
    - `sort` (str):
        Option to sort the submissions, default to `hot`.
        Available options: `hot`, `new`, `top`, `rising`
    - `t` (str):
        Type for sorting submissions by `top`, default to `day`.
        Available options: `hour`, `day`, `week`, `month`, `year`, `all`
    - `page_limit` (int):
        Set a request limit for pages to fetch. Disable this limit by passing `None`.
        Default to 4 (which will fetch 100 posts)
    - `req_delay` (int):
        Set delay between each page request. Set 0 to disable it. Default to 0.5.
    """
    log.debug(f"Fetching submissions and comments from subreddit *{subreddit_name}*")

    async with Client(**httpx_options) as client:
        token, dist = None, None

        while True:
            data = await client.get_posts(
                subreddit_name,
                sort=sort,
                t=t,
                after=token,
                dist=dist,
            )
            yield data

            # Check continue condition
            token, dist = data.get("token"), data.get("dist")
            if not token:
                break

            # Handle page limit
            if isinstance(page_limit, (int, float)):
                page_limit -= 1
                if page_limit < 1:
                    break

            # Apply delay
            await sleep(req_delay)


async def get_comments(
    submission_id: str,
    sort: Optional[str] = None,
    all_comments: bool = False,
    max_at_once: int = 8,
    max_per_second: int = 4,
    httpx_options: Dict[str, Any] = {},
):
    """
    Async Generator to get comments batch by batch.

    Returns an async generator that yields a list of comments. Use `async for` loop to handle the results.

    - `submission_id` (str):
        The Submission id (starts with `t3_`).
    - `sort` (str):
        Option to sort the comments of the submission, default to `None` (best).
        Available options: `top`, `new`, `controversial`, `old`, `qa`.
    - `all_comments` (bool):
        Set this to `True` to also get all nested comments. Default to `False`.
    - `max_at_once` (int):
        Limits the maximum number of concurrently requests for fetching all comments. Default to 8.
    - `max_per_second` (int):
        Limits the number of requests spawned per second. Default to 4.
    """
    log.debug(f"Fetching comments from submission *{submission_id}*")

    async with Client(**httpx_options) as client:
        raw_json = await client.raw_get_post_comments(submission_id, sort)
        yield list(raw_json["comments"].values())

        more_comments = raw_json["moreComments"].values()

        if all_comments and more_comments:
            while more_comments:
                reqs = [
                    partial(client.raw_get_more_comments, submission_id, mc["token"])
                    for mc in more_comments
                ]
                aggr_res = await run_all(
                    reqs,
                    max_at_once=max_at_once,
                    max_per_second=max_per_second,
                )

                # Yield comments
                yield [c for res in aggr_res for c in res["comments"].values()]

                # Extract more comments
                more_comments = [
                    comment
                    for res in aggr_res
                    for comment in res["moreComments"].values()
                ]


async def get_pushshift_posts(
    subreddit_name: str,
    start_desc: datetime = None,
    end_till: datetime = None,
    req_delay: int = 0.5,
    httpx_options: Dict[str, Any] = {},
):
    """
    Async Generator to get archived submissions by time range from pushshift.

    Returns an async generator that yields a list of posts. Use `async for` loop to handle the results.

    - `subreddit_name` (str):
        Name of the subreddit.
    - `start_desc` (`Optional[datetime]`):
        Provide `datetime` to get posts of a time range.
        Default to `None` to get from latest posts.
    - `end_till` (`Optional[datetime]`):
        Provide `datetime` to get posts of a time range.
        Default to `None` to get all existing posts.
    - `req_delay` (int):
        Set delay between each page request. Set 0 to disable it. Default to 0.5.
    """
    log.debug(f"Fetching pushshift submissions from subreddit *{subreddit_name}*")

    async with PushShiftAPI(**httpx_options) as client:
        before = get_timestamp(start_desc) if start_desc else None
        after = get_timestamp(end_till) if end_till else None

        while True:
            data = await client.get_posts(subreddit_name, before=before, after=after)
            yield data

            # Check continue condition
            if len(data) < client._DEFAULT_SIZE:
                break

            # Continue and apply delay
            before = data[-1]["created_utc"]
            await sleep(req_delay)



"""
This module defines gatered client to fetch information through Reddit webAPI (gateway.reddit.com).
"""
from typing import Any, Optional, TypeVar, Callable, Coroutine
from typing_extensions import TypeAlias
from functools import partial, partialmethod

from asyncio import sleep
from aiometer import run_all
from httpx import AsyncClient
import logging

T = TypeVar("T")
log = logging.getLogger(__name__)
_RequestType: TypeAlias = "Coroutine[None, None, Optional[T]]"


class Client:
    """
    The Client that interacts with Reddit webAPI and returns raw JSON as `dict`.
    [httpx options](https://www.python-httpx.org/api/#asyncclient) can be passed in when creating the client such as `proxies`.
    """

    _SUBREDDIT_SORT = "hot"
    _TOP_SORT_T = "day"
    _BASE_URL = "https://gateway.reddit.com/desktopapi/v1"
    _DEFAULT_HEADER = {
        "Accept": "*/*",
        "Accept-Encoding": "gzip",
        "Accept-Language": "en-US,en;q=0.5",
        "DNT": "1",
        "Origin": "https://www.reddit.com",
        "Referer": "https://www.reddit.com/",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:78.0) Gecko/20100101 Firefox/78.0",
    }
    _DEFAULT_PARAMS = {
        # rtj (rich text json) is excluded
        "redditWebClient": "web2x",
        "app": "web2x-client-production",
        "allow_over18": 1,
    }

    def __init__(self, **options: Any):
        """
        Initialize a `Client` object, where `options` are
        [httpx options](https://www.python-httpx.org/api/#asyncclient).
        """
        self._client: Optional[AsyncClient] = None
        self._x_reddit_loid: str = "0"
        self._x_reddit_session: str = "0"
        self.options = options

    async def __aenter__(self):
        self._client = AsyncClient(
            base_url=self._BASE_URL,
            headers=self._DEFAULT_HEADER,
            params=self._DEFAULT_PARAMS,
            **self.options,
        )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._client.aclose()

    async def _request(
        self,
        method: str,
        url_path: str,
        **kwargs: Any,
    ) -> Optional[Any]:
        kwargs["headers"] = kwargs.get("headers", {})
        kwargs["headers"].update(
            {
                "x-reddit-loid": self._x_reddit_loid,
                "x-reddit-session": self._x_reddit_session,
            }
        )

        for tries in range(1, 6):
            r = await self._client.request(method, url_path, **kwargs)
            log.debug(
                f"{method} {r.url} *{kwargs.get('data')}* has returned {r.status_code}"
            )

            # Successful returning text/json
            if 200 <= r.status_code < 300:
                if self._x_reddit_loid == "0":
                    self._x_reddit_loid = r.headers.get("x-reddit-loid", "")
                    self._x_reddit_session = r.headers.get("x-reddit-session", "")

                return r.json()

            # Rate limited (this is not tested)
            elif r.status_code == 429:
                log.warning("Rate limited!!")
                try:
                    await sleep(float(r.headers["X-Retry-After"]))
                except KeyError:
                    await sleep(30 * tries)
                continue

            # we've received a 500, 502 or 504, an unconditional retry
            elif r.status_code in {500, 502, 504}:
                await sleep(tries * 5)
                continue

            # then usual error cases - 403, 404 ...
            else:
                r.raise_for_status()

        # we've run out of retries, raise
        r.raise_for_status()

    _get: Callable[..., _RequestType] = partialmethod(_request, "GET")
    _post: Callable[..., _RequestType] = partialmethod(_request, "POST")

    # -- Request endpoints impl. --

    async def raw_get_more_comments(
        self,
        submission_id: str,
        token: str,
        **kwargs: Any,
    ):
        """
        Get `comments` from a submission provided with a generated cache `moreComment` token.
        Returns raw JSON.

        - `submission_id` (str):
            The Submission id (starts with `t3_`)
        - `token` (str):
            Token for more comments content.
        """
        payload = {"token": token}
        params = {"emotes_as_images": "true"}
        return await self._post(
            f"/morecomments/{submission_id}",
            params=params,
            json=payload,
            **kwargs,
        )

    async def raw_get_post_comments(
        self,
        submission_id: str,
        sort: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Get `post` and its `comments` + `moreComments` (including its related info).
        Return raw JSON.

        - `submission_id` (str):
            The Submission id (starts with `t3_`)
        - `sort` (str):
            Option to sort the comments of the submission, default to `None` (best).
            Available options: `top`, `new`, `controversial`, `old`, `qa`, `None`
        """
        params = {
            "emotes_as_images": "true",
            "hasSortParam": "false",
            "include_categories": "true",
            "onOtherDiscussions": "false",
        }
        if sort in {"top", "new", "controversial", "old", "qa"}:
            params["hasSortParam"] = "true"
            params["sort"] = sort

        return await self._get(
            f"/postcomments/{submission_id}",
            params=params,
            **kwargs,
        )

    async def raw_get_posts(
        self,
        subreddit_name: str,
        sort: Optional[str] = _SUBREDDIT_SORT,
        t: Optional[str] = _TOP_SORT_T,
        after: Optional[str] = None,
        dist: Optional[int] = None,
        **kwargs: Any,
    ):
        """
        Get submissions list from a subreddit. This also includes advertisement submissions.
        Returns raw JSON.

        - `subreddit_name` (str):
            The Subreddit name.
        - `sort` (str):
            Option to sort the submissions, default to `hot`.
            Available options: `hot`, `new`, `top`, `rising`
        - `t` (str):
            Type for sorting submissions by `top`, default to `day`.
            Available options: `hour`, `day`, `week`, `month`, `year`, `all`
        - `after` (str) and `dist` (str):
            Included in the returned raw JSON. Used for pagination.
        """
        params = {"layout": "classic"}
        # Check for sort
        if sort in {"hot", "new", "top", "rising"}:
            params["sort"] = sort
            if sort == "top":
                if t not in {"hour", "day", "week", "month", "year", "all"}:
                    t = self._TOP_SORT_T
                params["t"] = t
        # Check for pagnition
        if after and dist:
            params["after"] = after
            params["dist"] = dist

        return await self._get(f"/subreddits/{subreddit_name}", params=params, **kwargs)

    async def get_post_comments(
        self,
        submission_id: str,
        sort: Optional[str] = None,
        all_comments: bool = False,
        max_at_once: int = 8,
        max_per_second: int = 4,
        **kwargs: Any,
    ):
        """
        Get submission and its comments.
        If `all_comments` is `True`, it will fetch all the comments that are nested by reddit.

        Returns a dict with `post` as dict and its `comments` as list

        - `submission_id` (str):
            The Submission id (starts with `t3_`).
        - `sort` (str):
            Option to sort the comments of the submission, default to `None` (best).
            Available options: `top`, `new`, `controversial`, `old`, `qa`.
        - `all_comments` (bool):
            Set this to `True` to also get all nested comments. Default to `False`.
        - `max_at_once` (int):
            Limits the maximum number of concurrently requests for fetching all comments. Default to 8.
        - `max_per_second` (int):
            Limits the number of requests spawned per second. Default to 4.
        """
        raw_json = await self.raw_get_post_comments(submission_id, sort, **kwargs)

        post = raw_json["posts"].get(submission_id)
        comments = list(raw_json["comments"].values())

        more_comments = raw_json["moreComments"].values()

        if all_comments and more_comments:
            while more_comments:
                reqs = [
                    partial(
                        self.raw_get_more_comments,
                        submission_id,
                        more_c["token"],
                        **kwargs,
                    )
                    for more_c in more_comments
                ]
                aggr_res = await run_all(
                    reqs,
                    max_at_once=max_at_once,
                    max_per_second=max_per_second,
                )

                # Add comments to comments
                comments += [c for res in aggr_res for c in res["comments"].values()]

                # Extract more comments
                more_comments = [
                    more_c
                    for res in aggr_res
                    for more_c in res["moreComments"].values()
                ]

        return {"post": post, "comments": comments}

    async def get_posts(
        self,
        subreddit_name: str,
        sort: Optional[str] = "hot",
        t: Optional[str] = "day",
        after: Optional[str] = None,
        dist: Optional[int] = None,
        **kwargs: Any,
    ):
        """
        Get submissions list from a subreddit, with ads filtered.
        This provides flexibility for you to handle pagninations by yourself.

        Returns a dict with `subreddit` and its `posts` as list, as well as `token` and `dist` for paginations.

        - `subreddit_name` (str):
            The Subreddit name.
        - `sort` (str):
            Option to sort the submissions, default to `hot`
            Available options: `hot`, `new`, `top`, `rising`
        - `t` (str):
            Type for sorting submissions by `top`, default to `day`
            Available options: `hour`, `day`, `week`, `month`, `year`, `all`
        - `after` (str) and `dist` (str):
            Used for pagination.
        """
        raw_json = await self.raw_get_posts(
            subreddit_name, sort, t, after, dist, **kwargs
        )

        subreddit_info = {
            **(list(raw_json["subreddits"].values())[0]),
            **(list(raw_json["subredditAboutInfo"].values())[0]),
        }
        posts = [
            raw_json["posts"].get(i)
            for i in raw_json["postIds"]
            if not i.startswith("t3_z=")
        ]

        return {
            "subreddit": subreddit_info,
            "posts": posts,
            "sort": raw_json.get("listingSort"),
            "token": raw_json.get("token"),
            "dist": raw_json.get("dist"),
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gatered']

package_data = \
{'': ['*']}

install_requires = \
['aiometer>=0.3.0', 'httpx[http2]>=0.21.0']

setup_kwargs = {
    'name': 'gatered',
    'version': '1.1.0',
    'description': 'Reddit Gateway API Library',
    'long_description': "# GateRed\n\nA utils for interacting with Reddit Gateway API (Web API), w/ pushshift historical posts support.\n\n[![Latest Version](https://img.shields.io/pypi/v/gatered.svg)](https://pypi.python.org/pypi/gatered)\n[![Supported Python Versions](https://img.shields.io/pypi/pyversions/gatered)](https://pypi.python.org/pypi/gatered)\n[![CI](https://github.com/countertek/gatered/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/countertek/gatered/actions/workflows/ci.yml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![GitHub license](https://img.shields.io/github/license/countertek/gatered)](https://github.com/countertek/gatered/blob/main/LICENSE)\n\n**[Documentation](https://countertek.github.io/gatered)**\n\n## Why this library?\n\nAlthough Reddit has developer APIs and there are existing libraries (e.g. praw) to interact with reddit, there are still several drawbacks in terms of collecting data:\n\n- An API key is needed to collect data.\n- Rate limit is based on API keys.\n- Some fields are missing using developer APIs.\n\nTherefore, **gatered** exists just to counter this problem. It directly access Reddit's web API to get the whole information. No authentication is needed, and it supports proxy provided by [httpx](https://www.python-httpx.org/advanced/#http-proxying).\n\n## Install\n\nYou can install this library easily from pypi:\n\n```bash\n# with pip\npip install gatered\n\n# with poetry\npoetry add gatered\n```\n\n## Using\n\nThe library provides easy functions to get start fast:\n- `gatered.func.get_post_comments`\n- `gatered.func.get_posts`\n- `gatered.func.get_comments`\n- `gatered.func.get_pushshift_posts`\n\nAlternatively you can directly use `gatered.client.Client` and `gatered.pushshift.PushShiftAPI` classes as your base to implement your own logics.\n\nErrors can be handled by importing either `gatered.RequestError` or `gatered.HTTPStatusError`, see [httpx exceptions](https://www.python-httpx.org/exceptions/) to learn more.\n\nSee [`examples/`](https://github.com/countertek/gatered/tree/main/examples/) for more examples.\n",
    'author': 'Darwin',
    'author_email': 'darekaze@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CounterTek/gatered',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
This module defines PushShiftAPi client to fetch information from PushShift Reddit Archive.
"""
from typing import Any, Optional, TypeVar

from asyncio import sleep
from httpx import AsyncClient, Timeout
import logging

T = TypeVar("T")
log = logging.getLogger(__name__)


class PushShiftAPI:
    """
    The Client that interacts with the PushShift API and returns raw JSON as `dict`.
    [httpx options](https://www.python-httpx.org/api/#asyncclient) can be passed in when creating the client such as `proxies`.

    This acts as a helper to fetch past submissions based on time range (which is not supported by reddit).
    To get the comments, it would be better to get a list of `submission_id` from PushShift and use `gatered.func.get_comments` instead.
    """

    _BASE_URL = "https://api.pushshift.io"
    _DEFAULT_HEADER = {
        "Accept": "application/json",
        "Accept-Encoding": "gzip",
        "Accept-Language": "en-US,en;q=0.5",
        "Connection": "keep-alive",
        "DNT": "1",
        "Origin": "https://redditsearch.io/",
        "Referer": "https://redditsearch.io/",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:78.0) Gecko/20100101 Firefox/78.0",
    }
    _DEFAULT_OPTIONS = {"timeout": Timeout(10, read=30)}
    _DEFAULT_SORT = "desc"
    _DEFAULT_SIZE = 100

    def __init__(self, **options: Any):
        """
        Initialize a `PushShiftAPI` object, where `options` are
        [httpx options](https://www.python-httpx.org/api/#asyncclient).
        """
        self._client: Optional[AsyncClient] = None
        self._x_reddit_loid: str = "0"
        self._x_reddit_session: str = "0"
        self.options = {**self._DEFAULT_OPTIONS, **options}

    async def __aenter__(self):
        self._client = AsyncClient(
            base_url=self._BASE_URL,
            headers=self._DEFAULT_HEADER,
            **self.options,
        )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._client.aclose()

    async def _get(self, url_path: str, **kwargs: Any) -> Optional[Any]:
        for tries in range(1, 6):
            r = await self._client.get(url_path, **kwargs)
            log.debug(
                f"GET {r.url} *{kwargs.get('data')}* has returned {r.status_code}"
            )

            # Successful returning json
            if 200 <= r.status_code < 300:
                return r.json()

            # Rate limited and wait for a minute
            elif r.status_code == 429:
                log.warning("Rate limited!! Sleep for 60 secs")
                await sleep(60)
                continue

            # we've received a 500, 502 or 504, an unconditional retry
            elif r.status_code in {500, 502, 504}:
                await sleep(tries * 5)
                continue

            # then usual error cases - 403, 404 ...
            else:
                r.raise_for_status()

        # we've run out of retries, raise
        r.raise_for_status()

    async def get_posts(
        self,
        subreddit_name: str,
        before: int = None,
        after: int = None,
        sort: str = _DEFAULT_SORT,
        size: int = _DEFAULT_SIZE,
        **kwargs: Any,
    ):
        """
        Get a list of submissions from PushShift. Returns a list of PushShift archived submissions.

        - `subreddit_name` (str):
            The Subreddit name.
        - `before` (`Optional[int]`) and `after` (`Optional[int]`):
            Provide epoch time (without ms) to get posts from a time range.
            Default to `None` to get latest posts.
        - `sort` (str):
            Option to sort the submissions, default to `desc`
            Available options: `asc`, `desc`
        - `size` (int):
            Size of list to fetch. Default to maximum of 100.
        """
        res = await self._get(
            "/reddit/search/submission",
            params={
                "after": after,
                "before": before,
                "subreddit": subreddit_name,
                "sort": sort,
                "size": size,
            },
            **kwargs,
        )
        return res["data"]
